From 14410d2875c2f5894ee41735b7c0e9d2e1fdd41f Mon Sep 17 00:00:00 2001
From: Pavel Baturko <pabftk@gmail.com>
Date: Sat, 29 Sep 2012 01:57:06 +0000
Subject: [PATCH] USSD: Add decoding UCS2 USSD messages. UCS2 text is converted to UTF-8 and if terminal supports UTF-8 text will be printed, if not - hex representation of text will be printed.

---
 src/host/layer23/src/mobile/gsm480_ss.c |   66 ++++++++++++++++++++++--------
 1 files changed, 48 insertions(+), 18 deletions(-)

diff --git a/src/host/layer23/src/mobile/gsm480_ss.c b/src/host/layer23/src/mobile/gsm480_ss.c
index fda6288..9e1d4b9 100644
--- a/src/host/layer23/src/mobile/gsm480_ss.c
+++ b/src/host/layer23/src/mobile/gsm480_ss.c
@@ -35,6 +35,8 @@
 #include <osmocom/bb/mobile/vty.h>
 #include <osmocom/gsm/protocol/gsm_04_80.h>
 #include <osmocom/gsm/gsm48.h>
+#include <osmocom/gsm/gsm0411_utils.h>
+#include <osmocom/gsm/gsm_utils.h>
 
 static uint32_t new_callref = 0x80000001;
 
@@ -733,6 +735,10 @@ static int gsm480_rx_ussd(struct gsm_trans *trans, const uint8_t *data,
 	int i;
 	const uint8_t *tag_data;
 	int tag_len;
+	enum sms_alphabet alphabet;
+	uint8_t dcs;
+	uint8_t have_utf8_locale;
+	uint16_t utf8_text_length;
 
 	/* sequence tag */
 	if (parse_tag_asn1(data, len, &tag_data, &tag_len) < 0) {
@@ -755,10 +761,9 @@ static int gsm480_rx_ussd(struct gsm_trans *trans, const uint8_t *data,
 		LOGP(DSS, LOGL_NOTICE, "Expecting DSC tag\n");
 		return -EINVAL;
 	}
-	if (tag_data[0] != 0x0f) {
-		LOGP(DSS, LOGL_NOTICE, "DSC not 0x0f\n");
-		return -EINVAL;
-	}
+	
+	dcs = tag_data[0];
+	alphabet = gsm338_get_sms_alphabet(dcs);
 	len -= tag_data - data + tag_len;
 	data = tag_data + tag_len;
 
@@ -771,21 +776,46 @@ static int gsm480_rx_ussd(struct gsm_trans *trans, const uint8_t *data,
 		LOGP(DSS, LOGL_NOTICE, "Expecting text tag\n");
 		return -EINVAL;
 	}
-	num_chars = tag_len * 8 / 7;
-	/* Prevent a mobile-originated buffer-overrun! */
-	if (num_chars > sizeof(text) - 1)
-		num_chars = sizeof(text) - 1;
-	text[sizeof(text) - 1] = '\0';
-	gsm_7bit_decode(text, tag_data, num_chars);
-
-	for (i = 0; text[i]; i++) {
-		if (text[i] == '\r')
-			text[i] = '\n';
+	switch (alphabet) {
+	case DCS_7BIT_DEFAULT:
+		/* Default 7 bit charset */
+		num_chars = tag_len * 8 / 7;
+		/* Prevent a mobile-originated buffer-overrun! */
+		if (num_chars > sizeof(text) - 1)
+			num_chars = sizeof(text) - 1;
+		text[sizeof(text) - 1] = '\0';
+		gsm_7bit_decode(text, tag_data, num_chars);
+
+		for (i = 0; text[i]; i++) {
+			if (text[i] == '\r')
+				text[i] = '\n';
+		}
+		/* remove last CR, if exists */
+		if (text[0] && text[strlen(text) - 1] == '\n')
+			text[strlen(text) - 1] = '\0';
+		gsm480_ss_result(trans->ms, text, 0);
+		break;
+	case DCS_UCS2:
+		/* USC2 charset, see ISO/IEC 10646 */
+		num_chars = tag_len / 2;
+		utf8_text_length = ucs2_to_utf8_decode(text, sizeof(text), tag_data, num_chars);
+		
+		have_utf8_locale = check_utf8_locale();
+		if (have_utf8_locale) {
+			LOGP(DSS, LOGL_NOTICE, "Received USSD (UTF-8, text):\n%s\n",
+				text);
+			sprintf(text, "USSD was printed(decoded from UCS2 to UTF-8)");
+		} else {
+			LOGP(DSS, LOGL_NOTICE, "Received USSD (UTF-8, hex):\n%s\n", 
+				osmo_hexdump(text, utf8_text_length));
+			sprintf(text, "USSD was printed in hex(locale not UTF-8)");
+		}
+		gsm480_ss_result(trans->ms, text, 0);
+		break;
+	default:
+		LOGP(DSS, LOGL_NOTICE, "DSC <%d> unknown\n", dcs);
+		return -EINVAL;
 	}
-	/* remove last CR, if exists */
-	if (text[0] && text[strlen(text) - 1] == '\n')
-		text[strlen(text) - 1] = '\0';
-	gsm480_ss_result(trans->ms, text, 0);
 
 	return 0;
 }
-- 
1.7.2.5

